import django
from django.db import models
from django.conf import settings
from prices import Price

from . import forms

if django.VERSION < (1, 8):
    BaseField = models.SubfieldBase('BaseField', (models.DecimalField,), {})
else:
    BaseField = models.DecimalField


class Creator(object):
    """
    A placeholder class that provides a way to set the attribute on the model.
    """
    def __init__(self, field):
        self.field = field

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self
        return obj.__dict__[self.field.name]

    def __set__(self, obj, value):
        obj.__dict__[self.field.name] = self.field.to_python(value)


class PriceField(BaseField):

    description = 'A field which stores a price.'

    def __init__(self, verbose_name=None, currency=None, **kwargs):
        self.currency = currency
        super(PriceField, self).__init__(verbose_name, **kwargs)

    def contribute_to_class(self, cls, name, **kwargs):
        super(PriceField, self).contribute_to_class(cls, name, **kwargs)
        setattr(cls, self.name, Creator(self))

    def from_db_value(self, value, expression, connection, context):
        return self.to_python(value)

    def to_python(self, value):
        if isinstance(value, Price):
            if value.currency != self.currency:
                raise ValueError('Invalid currency: %r (expected %r)' % (
                    value.currency, self.currency))
            return value
        value = super(PriceField, self).to_python(value)
        if value is None:
            return value
        return Price(value, currency=self.currency)

    def get_prep_value(self, value):
        value = self.to_python(value)
        if value is not None:
            value = value.net
        return value

    def get_db_prep_save(self, value, connection):
        value = self.get_prep_value(value)
        if django.VERSION < (1, 9):
            db_value = connection.ops.value_to_db_decimal
        else:
            db_value = connection.ops.adapt_decimalfield_value
        return db_value(value, self.max_digits, self.decimal_places)

    def value_to_string(self, obj):
        value = self.value_from_object(obj)
        if value is not None:
            return value.net
        return super(PriceField, self).value_to_string(value)

    def formfield(self, **kwargs):
        defaults = {'currency': self.currency,
                    'form_class': forms.PriceField}
        defaults.update(kwargs)
        return super(PriceField, self).formfield(**defaults)

    def get_default(self):
        default = super(PriceField, self).get_default()
        return self.to_python(default)

    def deconstruct(self):
        name, path, args, kwargs = super(PriceField, self).deconstruct()
        kwargs['currency'] = self.currency
        return name, path, args, kwargs

if 'south' in settings.INSTALLED_APPS:
    from south.modelsinspector import add_introspection_rules
    rules = [
        ((PriceField,), [], {
            'currency': ('currency', {})
        })
    ]
    add_introspection_rules(rules, ['^django_prices\.models'])



from django import forms, VERSION as DJANGO_VERSION
from prices import Price

from .widgets import PriceInput

__all__ = ('PriceField', 'PriceInput')


class PriceField(forms.DecimalField):

    def __init__(self, currency, widget=PriceInput, *args, **kwargs):
        self.currency = currency
        if isinstance(widget, type):
            widget = widget(currency=self.currency, attrs={
                'type': 'number', 'step': 'any'})
        super(PriceField, self).__init__(*args, widget=widget, **kwargs)

    def to_python(self, value):
        value = super(PriceField, self).to_python(value)
        if value is None:
            return value
        return Price(value, currency=self.currency)

    def validate(self, value):
        if value is None:
            super(PriceField, self).validate(value)
        else:
            if not isinstance(value, Price):
                raise Exception('%r is not a valid price' % (value,))
            if value.currency != self.currency:
                raise forms.ValidationError(
                    'Invalid currency: %r (expected %r)' % (
                        value.currency, self.currency))
            super(PriceField, self).validate(value.net)

    def run_validators(self, value):
        if isinstance(value, Price):
            value = value.net
        return super(PriceField, self).run_validators(value)

    def has_changed(self, initial, data):
        if not isinstance(initial, Price):
            initial = self.to_python(initial)
        return super(PriceField, self).has_changed(initial, data)

    if DJANGO_VERSION < (1, 8):
        def _has_changed(self, initial, data):
            if not isinstance(initial, Price):
                initial = self.to_python(initial)
            return super(PriceField, self)._has_changed(initial, data)



from __future__ import unicode_literals

from django import forms, VERSION as DJANGO_VERSION
from django.template.loader import render_to_string

from prices import Price

__all__ = ['PriceInput']


class PriceInput(forms.TextInput):
    template = 'prices/widget.html'
    input_type = 'number'

    def __init__(self, currency, *args, **kwargs):
        self.currency = currency
        super(PriceInput, self).__init__(*args, **kwargs)

    def _format_value(self, value):
        if isinstance(value, Price):
            value = value.net
        return value

    if DJANGO_VERSION < (1, 6):
        def _has_changed(self, initial, data):
            if isinstance(initial, Price):
                initial = initial.net
            return super(PriceInput, self)._has_changed(initial, data)

    def render(self, name, value, attrs=None):
        widget = super(PriceInput, self).render(name, value, attrs=attrs)
        return render_to_string(self.template, {'widget': widget,
                                                'value': value,
                                                'currency': self.currency})



from decimal import Decimal

import django
from django import forms as django_forms
from django.db import connection, models
from prices import Price
import pytest

from . import forms
from . import widgets
from .models import PriceField


@pytest.fixture(scope='module')
def django_setup():
    if django.VERSION >= (1, 7, 0):
        django.setup()


@pytest.fixture(scope='module')
def test_model(django_setup):
    class TestModel(models.Model):
        price = PriceField(currency='BTC', default='5', max_digits=9,
                           decimal_places=2)
    return TestModel


@pytest.fixture(scope='module')
def test_form(django_setup):
    class PriceForm(django_forms.Form):
        price = forms.PriceField(currency='BTC')
    return PriceForm

@pytest.fixture(scope='module')
def test_form_price_not_required(django_setup):
    class PriceForm(django_forms.Form):
        price = forms.PriceField(currency='BTC', required=False)
    return PriceForm


def test_init():
    field = PriceField(
        currency='BTC', default='5', max_digits=9, decimal_places=2)
    assert field.get_default() == Price(5, currency='BTC')


def test_get_prep_value():
    field = PriceField('price', currency='BTC', default='5', max_digits=9,
                       decimal_places=2)
    assert field.get_prep_value(Price(5, currency='BTC')) == Decimal(5)


def test_get_db_prep_save():
    field = PriceField('price', currency='BTC', default='5', max_digits=9,
                       decimal_places=2)
    value = field.get_db_prep_save(Price(5, currency='BTC'), connection)
    assert value == '5.00'


def test_value_to_string(test_model):
    instance = test_model(price=30)
    field = instance._meta.get_field('price')
    assert field.value_to_string(instance) == Decimal('30')


def test_from_db_value():
    field = PriceField('price', currency='BTC', default='5', max_digits=9,
                       decimal_places=2)
    assert field.from_db_value(7, None, None, None) == Price(7, currency='BTC')


def test_from_db_value_handles_none():
    field = PriceField('price', currency='BTC', default='5', max_digits=9,
                       decimal_places=2)
    assert field.from_db_value(None, None, None, None) == None


def test_from_db_value_checks_currency():
    field = PriceField('price', currency='BTC', default='5', max_digits=9,
                       decimal_places=2)
    invalid = Price(1, currency='USD')
    with pytest.raises(ValueError):
        field.from_db_value(invalid, None, None, None)


def test_formfield():
    field = PriceField('price', currency='BTC', default='5', max_digits=9,
                       decimal_places=2)
    form_field = field.formfield()
    assert isinstance(form_field, forms.PriceField)
    assert form_field.currency == 'BTC'
    assert isinstance(form_field.widget, widgets.PriceInput)


@pytest.mark.parametrize("data,initial,expected_result", [
    ('5', Price(5, currency='BTC'), False),
    ('5', Price(10, currency='BTC'), True),
    ('5', '5', False),
    ('5', '10', True),
    ('5', None, True),
    (None, Price(5, currency='BTC'), True),
    (None, '5', True),
    (None, None, False)
])
def test_form_changed_data(test_form, data, initial, expected_result):
    form = test_form(data={'price': data}, initial={'price': initial})
    assert bool(form.changed_data) == expected_result


def test_render(django_setup):
    widget = widgets.PriceInput('BTC', attrs={'type': 'number'})
    result = widget.render('price', 5, attrs={'foo': 'bar'})
    attrs = [
        'foo="bar"', 'name="price"', 'type="number"', 'value="5"', 'BTC']
    for attr in attrs:
        assert attr in result


def test_instance_values(test_model):
    instance = test_model(price=25)
    assert instance.price.net == 25


def test_field_passes_all_validations(test_form):
    form = test_form(data={'price': '20'})
    form.full_clean()
    assert form.errors == {}


def test_field_passes_none_validation(test_form_price_not_required):
   form = test_form_price_not_required(data={'price': None})
   form.full_clean()
   assert form.errors == {}



from babeldjango.templatetags.babel import currencyfmt
from django import template

register = template.Library()


@register.simple_tag
def gross(price):
    return currencyfmt(price.gross, price.currency)


@register.simple_tag
def net(price):
    return currencyfmt(price.net, price.currency)


@register.simple_tag
def tax(price):
    return currencyfmt(price.tax, price.currency)



#! /usr/bin/env python
import os
from setuptools import setup
from setuptools.command.test import test as TestCommand
import sys

os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'test_settings')

CLASSIFIERS = [
    'Environment :: Web Environment',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: BSD License',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Topic :: Internet :: WWW/HTTP',
    'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    'Topic :: Software Development :: Libraries :: Application Frameworks',
    'Topic :: Software Development :: Libraries :: Python Modules']


class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]
    test_args = []

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        #import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)

setup(
    name='django-prices',
    author='Mirumee Software',
    author_email='hello@mirumee.com',
    description='Django fields for the prices module',
    license='BSD',
    version='0.4.7',
    url='https://github.com/mirumee/django-prices',
    packages=['django_prices', 'django_prices.templatetags'],
    modules=['test_settings'],
    include_package_data=True,
    classifiers=CLASSIFIERS,
    install_requires=['BabelDjango', 'django', 'prices>=0.5,<0.6a0'],
    tests_require=['pytest'],
    platforms=['any'],
    test_suite='django_prices.tests',
    cmdclass={
        'test': PyTest},
    zip_safe=False)
